-- Foot profile

api_version = 2

Set = require('lib/set')
Sequence = require('lib/sequence')
Handlers = require("lib/way_handlers")
find_access_tag = require("lib/access").find_access_tag

function setup()
  local walking_speed = 3
  return {
    properties = {
      weight_name                   = 'duration',
      max_speed_for_map_matching    = 40/3.6, -- kmph -> m/s
      call_tagless_node_function    = false,
      traffic_light_penalty         = 2,
      u_turn_penalty                = 2,
      continue_straight_at_waypoint = false,
      use_turn_restrictions         = false,
      fallback_names                = true, -- commenta qua se non funziona
      obey_oneway                   = true -- come sopra
    },

    default_mode            = mode.walking,
    default_speed           = walking_speed,
    oneway_handling         = 'specific',     -- respect 'oneway:foot' but not 'oneway'

    barrier_whitelist = {
      '',
      'bollard',
      'entrance',
      'border_control',
      'toll_booth',
      'sally_port',
      'gate',
      'no',
      'block',
      'kerb'
    },

    access_tag_whitelist = {
      'yes',
      'foot',
      'permissive',
      'designated'
    },

    access_tag_blacklist = {
      'no',
      'private',
      'agricultural',
      'forestry'
    },

    access_tag_restricted = {
      'destination',
      'delivery'
    },

    access_tags_hierachy = Sequence {
      "foot",
      "access"
    },

    service_tag_restricted = {
      'parking_aisle'
    },

    ignore_in_grid = {
      'ferry'
    },
    
    restriction_exception_tags = {
      'foot'
    },

    -- list of suffixes to suppress in name change instructions
    suffix_list = Set {
      'N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW', 'North', 'South', 'West', 'East'
    },

    avoid = Set {
      'impassable'
    },

    speeds = Sequence {
      highway = {
        primary = 0.1,
        primary_link = 0.1,
        secondary = 0.2,
        secondary_link = 0.2,
        tertiary = 0.3,
        tertiary_link = 0.3,
        unclassified = 0.5,
        residential = 1,
        sidewalk = walking_speed,
        road = walking_speed,
        living_street = walking_speed,
        service = walking_speed*0.75,
        track = walking_speed,
        path = walking_speed,
        steps = walking_speed*0.1, -- have to take it if it has a ramp
        pedestrian = walking_speed,
        platform = walking_speed,
        footway = walking_speed,
        pier = walking_speed,
        default = walking_speed
      },

      railway = {
        platform        = walking_speed
      },

      amenity = {
        parking = 0,
        parking_entrance = walking_speed
      },

      man_made = {
        pier            = walking_speed
      },

      leisure = {
        track           = walking_speed
      }
    },

    route_speeds = {
      ferry = 5
    },

    railway_speeds = { -- se non funziona commenta questo
      funicular = 0.1
    },

    bridge_speeds = {
    },

    surface_speeds = {
      fine_gravel =   walking_speed*0.25,
      gravel =        walking_speed*0.05,
      pebblestone =   0,
      cobblestone =   walking_speed*0.5,
      ['cobblestone:flattened'] =   walking_speed*0.5,
      paving_stones =   walking_speed*0.9,
      mud =           0,
      sand =          0,
      grass =         0,
      ground =         0,
      earth =         0,
      grass_paver =         0,
      unpaved =       walking_speed*0.1
    },

    tracktype_speeds = {
    },

    smoothness_speeds = {
      excellent =         1,
      good =         1,
      intermediate =         0.5,
      bad =         0,
      very_bad =         0,
      horrible =         0,
      impassable =         0,
    }
  }
end

-- Sono Arrivato QUI

function process_node(profile, node, result)
  -- parse access and barrier tags
  local access = find_access_tag(node, profile.access_tags_hierarchy)
  if access then
    if profile.access_tag_blacklist[access] then
      result.barrier = true
    end
  else
    local barrier = node:get_value_by_key("barrier")
    if barrier then
      --  make an exception for rising bollard barriers
      local bollard = node:get_value_by_key("bollard")
      local rising_bollard = bollard and "rising" == bollard

      if profile.barrier_blacklist[barrier] and not rising_bollard then
        result.barrier = true
      end
    end
  end

  -- check if node is a traffic light
  local tag = node:get_value_by_key("highway")
  if "traffic_signals" == tag then
    result.traffic_lights = true
  end
    -- wheelchair tag overrides all
  if wheelchair and (wheelchair == "yes" or wheelchair == "designated" ) then
    result.barrier = false
    return 1
  end
  if wheelchair and wheelchair ~= "" and barrier and barrier ~= "" and barrier ~= "no" then -- all other values (limited is only for wheelchair with help) on barriers
      result.barrier = true
      return 1
  end
  if wheelchair and wheelchair ~= "" and humps and humps ~= "" and humps ~= "no" then -- all other values (limited is only for wheelchair with help) on barriers
      result.barrier = true
      return 1
  end
  
  -- if step tagged on a node, they are a barrier
  if highway and highway == "steps" and not (wheelchair_ramp and wheelchair_ramp == "yes") then
    result.barrier = true
    return 1
  end
  -- nodes with crossing=no are a no-cross
  if crossing and crossing == "no" then
    result.barrier = true
    return 1
  end

  if width and width ~= "" then
      if width:match("^[0-9.]+%s?m?$") then
          lwidth = tonumber(width:match("^[0-9.]+"))
          if lwidth ~= nil then
              if lwidth >= minwidth then
                  result.barrier = false
              else
                  result.barrier = true
                  return 1
              end
          end
      elseif width:match("^[0-9.]+%s?cm$") then
          lwidth = tonumber(width:match("^[0-9.]+"))
          if lwidth ~= nil then
              if lwidth >= minwidth*100 then
                  result.barrier = false
              else
                  result.barrier = true
                  return 1
              end
          end
      end
  end

  -- barrier: kerb:height is the crucial value
  -- is it only [0-9.], then it is in m
  -- is the unit cm
  if kerbheight and kerbheight ~= "" then
      if kerbheight:match("^[0-9.]+%s?m?$") then
          lheight = tonumber(kerbheight:match("^[0-9.]+"))
          if lheight ~= nil then
              if lheight <= maxkerbheight then
                  result.barrier = false
              else
                  result.barrier = true
              end
              return 1
          end
      elseif kerbheight:match("^[0-9.]+%s?cm$") then
          lheight = tonumber(kerbheight:match("^[0-9.]+"))
          if lheight ~= nil then
              if lheight <= maxkerbheight*100 then
                  result.barrier = false
              else
                  result.barrier = true
              end
              return 1
          end
      end
      -- couldn't be parsed, ignore
  end
  -- no kerbheight set or not readable
  if kerbtype and kerbtype ~= "" then
      if kerbtype == "flush" or kerbtype == "lowered" or kerbtype == "no" then
          result.barrier = false
      else
          result.barrier = true
      end
      return 1
  end

  -- parse access and barrier tags
  if access and access ~= "" then
    if access_tag_blacklist[access] then
      result.barrier = true
    else
      result.barrier = false
    end
  elseif barrier and barrier ~= "" then
    if barrier_whitelist[barrier] then
      result.barrier = false
    else
      result.barrier = true
    end
  end
  return 1
end

-- main entry point for processsing a way
function process_way(profile, way, result)
  -- the intial filtering of ways based on presence of tags
  -- affects processing times significantly, because all ways
  -- have to be checked.
  -- to increase performance, prefetching and intial tag check
  -- is done in directly instead of via a handler.

  -- in general we should  try to abort as soon as
  -- possible if the way is not routable, to avoid doing
  -- unnecessary work. this implies we should check things that
  -- commonly forbids access early, and handle edge cases later.

  -- data table for storing intermediate values during processing
  local data = {
    -- prefetch tags
    highway = way:get_value_by_key("highway"),
    wheelchair_ramp = way:get_value_by_key("ramp:wheelchair"),
    leisure = way:get_value_by_key("leisure"),
    route = way:get_value_by_key("route"),
    man_made = way:get_value_by_key("man_made"),
    railway = way:get_value_by_key("railway"),
    amenity = way:get_value_by_key("amenity"),
    public_transport = way:get_value_by_key("public_transport"),
    wheelchair = way:get_value_by_key("wheelchair"),
    width = way:get_value_by_key("width"),
    incline = way:get_value_by_key("incline"),
    incline_across = way:get_value_by_key("incline:across"),
    sidewalk = way:get_value_by_key("sidewalk"),
    name = way:get_value_by_key("name"),
    footway_type = way:get_value_by_key("footway"),
    ref = way:get_value_by_key("ref"),
    junction = way:get_value_by_key("junction"),
    onewayClass = way:get_value_by_key("oneway"),
    oneway_cycle = way:get_value_by_key("oneway:bicycle"),
    oneway_foot = way:get_value_by_key("oneway:foot"),
    oneway_wheelchair = way:get_value_by_key("oneway:wheelchair"),
    cycleway = way:get_value_by_key("cycleway"),
    cycleway_right = way:get_value_by_key("cycleway:right"),
    cycleway_left = way:get_value_by_key("cycleway:left"),
    duration  = way:get_value_by_key("duration"),
    service  = way:get_value_by_key("service"),
    area = way:get_value_by_key("area"),
    foot = way:get_value_by_key("foot"),
    surface = way:get_value_by_key("surface"),
    smoothness = way:get_value_by_key("smoothness")
  }

  -- perform an quick initial check and abort if the way is
  -- obviously not routable. here we require at least one
  -- of the prefetched tags to be present, ie. the data table
  -- cannot be empty
  if next(data) == nil then     -- is the data table empty?
    return
  end

  if (not highway or highway == '') and
    (not leisure or leisure == '') and
    (not route or route == '') and
    (not railway or railway=='') and
    (not amenity or amenity=='') and
    (not man_made or man_made=='') and
    (not public_transport or public_transport=='')
    then
    return
  end

  if highway=='construction' then
    return
  end

  if wheelchair and (wheelchair == "no" or wheelchair == "limited") then
    return
  end

  if highway and highway == "steps" and not ( (wheelchair_ramp and  wheelchair_ramp == "yes") or (wheelchair and wheelchair == "yes") ) then
    return
  end

  -- access
  local access = find_access_tag(way, access_tags_hierachy)
  if access_tag_blacklist[access] then
    return
  end

  result.forward_mode = mode.walking
  result.backward_mode = mode.walking

  if ref and "" ~= ref and name and "" ~= name then
    result.name = name .. ' / ' .. ref
    elseif ref and "" ~= ref then
      result.name = ref
  elseif name and "" ~= name then
    result.name = name
  elseif footway_type and "" ~= footway_type then
      result.name = footway_type
  elseif highway and fallback_names then
    result.name = "{highway:"..highway.."}"  -- if no name exists, use way type
                                            -- this encoding scheme is excepted to be a temporary solution
  end

    -- roundabouts
  if "roundabout" == junction then
    result.roundabout = true
  end

    -- speed
  if route_speeds[route] then
    -- ferries (doesn't cover routes tagged using relations)
    result.ignore_in_grid = true
    if duration and durationIsValid(duration) then
      result.duration = math.max( 1, parseDuration(duration) )
    else
      result.forward_speed = route_speeds[route]
      result.backward_speed = route_speeds[route]
    end
    result.forward_mode = mode_ferry
    result.backward_mode = mode_ferry
  elseif railway and platform_speeds[railway] then
    -- railway platforms (old tagging scheme)
    result.forward_speed = platform_speeds[railway]
    result.backward_speed = platform_speeds[railway]
  elseif platform_speeds[public_transport] then
    -- public_transport platforms (new tagging platform)
    result.forward_speed = platform_speeds[public_transport]
    result.backward_speed = platform_speeds[public_transport]
  elseif amenity and amenity_speeds[amenity] then
    -- parking areas
    result.forward_speed = amenity_speeds[amenity]
    result.backward_speed = amenity_speeds[amenity]
  elseif leisure and leisure_speeds[leisure] then
    -- running tracks
    result.forward_speed = leisure_speeds[leisure]
    result.backward_speed = leisure_speeds[leisure]
  elseif speeds[highway] then
    -- regular ways
    result.forward_speed = speeds[highway]
    result.backward_speed = speeds[highway]
  elseif railway and railway_speeds[railway] then
    -- special railway like funicular
    result.forward_speed = railway_speeds[railway]
    result.backward_speed = railway_speeds[railway]
  elseif access and access_tag_whitelist[access] then
      -- unknown way, but valid access tag
    result.forward_speed = walking_speed
    result.backward_speed = walking_speed
  end

  --if has sidewalk tagged as implicit, set speed to normal
  if sidewalk and (sidewalk == "both" or sidewalk == "left" or sidewalk == "right") then
      result.forward_speed = speeds["sidewalk"]
      result.backward_speed = speeds["sidewalk"]
  end

  -- oneway
  if onewayClass and onewayClass ~= "" then
    if oneway_wheelchair and oneway_wheelchair == "no"  then 
      -- do nothing
    elseif oneway_foot and oneway_foot == "no" then
      -- do nothing
    elseif highway == "cycleway" and foot and (foot == "yes" or foot == "designated" or foot == "permissive" ) then 
      -- do nothing
    elseif sidewalk and (sidewalk == "both" or sidewalk == "left" or sidewalk == "right") then
      -- do nothing
    elseif cycleway and (cycleway == "opposite" or cycleway == "opposite_lane" or cycleway == "opposite_track" ) then
      -- do nothing
    elseif oneway_cycle and oneway_cycle == "no" then
      -- do nothing

    else -- catch all remaining-block
      if onewayClass == "yes" or onewayClass == "1" or onewayClass == "true" then
        result.backward_mode = 0
      elseif onewayClass == "-1" then
        result.forward_mode = 0
      end
    end

  end

  -- surfaces
  if surface then
    surface_speed = surface_speeds[surface]
    if surface_speed then
      result.forward_speed = math.min(result.forward_speed, surface_speed)
      result.backward_speed  = math.min(result.backward_speed, surface_speed)
    end
  end

  if not wheelchair or not ( wheelchair and (wheelchair == "yes" or wheelchair == "designated") ) then
      -- set speed to 0 for too much inclined
      if incline and incline ~= "" then
          if incline:match("^-?[0-9.]+%s?%%?$") then
              lincline = tonumber(incline:match("[0-9.]+"))
              if lincline ~= nil and lincline > maxincline then
                  result.forward_speed = 0
                  result.backward_speed = 0
              end
          end
          if (incline == "up" or incline == "down") and not (wheelchair_ramp and wheelchair_ramp == "yes") then -- is only set when not known exactly - and most people only recognize or see inclines > 3% as relevant
              result.forward_speed = 0
              result.backward_speed = 0
          end
      end
      if incline_across and incline_across ~= "" then
          if incline_across:match("^-?[0-9.]+%s?%%?$") then
              lincline = tonumber(incline_across:match("[0-9.]+"))
              if lincline ~= nil and lincline > maxincline_across then
                  result.forward_speed = 0
                  result.backward_speed = 0
              end
          end
      end

      -- set speed to 0 for too narrow:
      if width and width ~= "" then
          if width:match("^[0-9.]+%s?m?$") then
              lwidth = tonumber(width:match("^[0-9.]+"))
              if lwidth ~= nil and lwidth < minwidth then
                  result.forward_speed = 0
                  result.backward_speed = 0
              end
          elseif width:match("^[0-9.]+%s?cm$") then
              lwidth = tonumber(width:match("^[0-9.]+"))
              if lwidth ~= nil and lwidth < minwidth*100 then
                  result.forward_speed = 0
                  result.backward_speed = 0
              end
          end
      end
  end

  local handlers = Sequence {
    -- set the default mode for this profile. if can be changed later
    -- in case it turns we're e.g. on a ferry
    WayHandlers.default_mode,

    -- check various tags that could indicate that the way is not
    -- routable. this includes things like status=impassable,
    -- toll=yes and oneway=reversible
    WayHandlers.blocked_ways,

    -- determine access status by checking our hierarchy of
    -- access tags, e.g: motorcar, motor_vehicle, vehicle
    WayHandlers.access,

    -- check whether forward/backward directons are routable
    WayHandlers.oneway,

    -- check whether forward/backward directons are routable
    WayHandlers.destinations,

    -- check whether we're using a special transport mode
    WayHandlers.ferries,
    WayHandlers.movables,

    -- compute speed taking into account way type, maxspeed tags, etc.
    WayHandlers.speed,
    WayHandlers.surface,

    -- handle turn lanes and road classification, used for guidance
    WayHandlers.classification,

    -- handle various other flags
    WayHandlers.roundabouts,
    WayHandlers.startpoint,

    -- set name, ref and pronunciation
    WayHandlers.names,

    -- set weight properties of the way
    WayHandlers.weights
  }

  WayHandlers.run(profile, way, result, data, handlers)
end

function process_turn (profile, turn)
  turn.duration = 0.

  if turn.direction_modifier == direction_modifier.u_turn then
     turn.duration = turn.duration + profile.properties.u_turn_penalty
  end

  if turn.has_traffic_light then
     turn.duration = profile.properties.traffic_light_penalty
  end
  if profile.properties.weight_name == 'routability' then
      -- penalize turns from non-local access only segments onto local access only tags
      if not turn.source_restricted and turn.target_restricted then
          turn.weight = turn.weight + 3000
      end
  end
end

return {
  setup = setup,
  process_way =  process_way,
  process_node = process_node,
  process_turn = process_turn
}